#!/bin/bash

# This script looks for a GMP installation in a standard location.
# If a single GMP installation is found, it prints out the full path of
# the (static) library and returns with exit code 0.
# If none is found or several are found, it prints out an error message
# and returns with a non-zero exit code.

##################################################################
# Use find to search through various standard directories.
# NB look through all directories, even if a GMP has already been found.

# List of directories under which libgmp.a is normally found.
STD_GMP_LIBDIRS="/usr/lib64:/usr/local:/usr/lib:/sw/lib"
IFS=":"
LIBGMPPATHS=libgmp-paths
/bin/rm -rf $LIBGMPPATHS
for directory in $STD_GMP_LIBDIRS
do
  if [ -d "$directory" ]
  then
    find "$directory" -name libgmp.a -print >> $LIBGMPPATHS
  fi
done
/bin/ls -t `cat $LIBGMPPATHS` > $LIBGMPPATHS.ordered
/bin/rm $LIBGMPPATHS
if [ \! -s $LIBGMPPATHS.ordered ]
then
  # Did not find any plausible GMP installation, so return empty handed.
  echo "No GMP installation found; looked inside $STD_GMP_LIBDIRS"
  /bin/rm $LIBGMPPATHS.ordered
  exit 1
fi

if [ `wc -l < $LIBGMPPATHS.ordered` -ne 1 ]
then
  echo "Found multiple GMP libraries:"
  cat $LIBGMPPATHS.ordered
  /bin/rm $LIBGMPPATHS.ordered
  exit 2
fi

# We have found a single file called libgmp.a; do a couple of quick
# sanity checks before decalring our serach successful...
GMP_LIB=`head -1 $LIBGMPPATHS.ordered`
/bin/rm $LIBGMPPATHS.ordered
GMP_DIR=`dirname "$GMP_LIB"`
GMP_HDR="$GMP_DIR/../include/gmp.h"
if [ -f "$GMP_LIB" -a -r "$GMP_LIB" -a -f "$GMP_HDR" -a -r "$GMP_HDR" ]
then
  # We've found a plausible libgmp.a and (probably) its corresponding gmp.h.
  echo "$GMP_LIB"
  exit 0
fi

# Get here probably only if there's a damaged GMP installation.
echo "Trouble reading file $GMP_LIB or its header file."
exit 3



##################################################################
# Here is our list of standard locations for the static library.
# We just use the first we find (check /usr/local/lib *before*
# /usr/lib expecting there might be a more recent "local" version).

STD_GMP="/usr/lib64/libgmp.a:/usr/local/gmp-4.2.1-static/lib/libgmp.a:/usr/local/gmp-4.2-static/lib/libgmp.a:/usr/local/gmp-4.1.4-static/lib/libgmp.a:/usr/local/lib/libgmp.a:/usr/lib/libgmp.a:/sw/lib/libgmp.a"

# Loop through list of candidates: accept candidate if both library
# and header file exist and are readable.
IFS=":"
for candidate in $STD_GMP
do
  GMP_DIR=`dirname $candidate`
  GMP_HDR="$GMP_DIR/../include/gmp.h"
  if [ -f "$candidate" -a -r "$candidate" -a -f "$GMP_HDR" -a -r "$GMP_HDR" ]
  then
    # Candidate is acceptable; print it out and exit.
    echo "$candidate"
    exit
  fi
done

# Did not find any plausible GMP installation, so return empty handed.
