//   Copyright (c)  1997-2006  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License (version 2)
//   as published by the Free Software Foundation.  A copy of the full
//   licence may be found in the file COPYING in this directory.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoA; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef QUADRATIC

#include <stddef.h>
#include <stdio.h>
#include "logi.h"
#include "DUPZfactor_lift.h"
#include "DUPIsubtract_product.h"
#include "jalloc.h"
#include "jaaerror.h"
#include "DUPI_DUPFF.h"
#include "DUPZfactor_info.h"
#include "DUPZpadic_expand.h"


/***************************************************************************/


void solve_for_correction(DUPFF *gdelta, const DUPFF E, const DUPFF g0, const DUPFF grecip)
{
  DUPFF junk, tmp;
  int dg0, dE;

  dg0 = DUPFFdeg(g0);
  dE = DUPFFdeg(E);
  junk = DUPFFnew(dE + dg0);  /* these are all a bit too big */
  tmp = DUPFFnew(2*dg0);      /*                             */
  *gdelta = DUPFFnew(dg0);    /*                             */

  DUPFFdiv4(junk, tmp, E, g0);
  DUPFFmul3(tmp, tmp, grecip);
  DUPFFdiv4(junk, *gdelta, tmp, g0);

  DUPFFfree(tmp);
  DUPFFfree(junk);
}

/***************************************************************************/


DUPZfactor_lifter DUPZfactor_lifter_ctor(DUPFF *g, DUPFF *h, DUPZfactor_lifter g_lifter, DUPZfactor_lifter h_lifter, int rmax)
{
  DUPZfactor_lifter ans;
  DUPFF grecip, hrecip, tmp;
  int df;
  
  ans = (DUPZfactor_lifter)MALLOC(sizeof(struct DUPZfactor_lift_struct));
  ans->r = 1;
  ans->g = (DUPFF*)MALLOC(rmax*sizeof(DUPFF));
  if (g_lifter) ans->g[0] = DUPFFmul(g_lifter->g[0], g_lifter->h[0]);
  else ans->g[0] = DUPFFcopy(*g);
  ans->h = (DUPFF*)MALLOC(rmax*sizeof(DUPFF));
  if (h_lifter) ans->h[0] = DUPFFmul(h_lifter->g[0], h_lifter->h[0]);
  else ans->h[0] = DUPFFcopy(*h);
  df = DUPFFdeg(ans->g[0]) + DUPFFdeg(ans->h[0]);
  tmp = DUPFFexgcd(&grecip, &hrecip, ans->g[0], ans->h[0]);
  if (DUPFFdeg(tmp) != 0) { JERROR(JERROR_HENSEL); exit(1); }
  /* since the gcd is not monic we have to do this... */
  DUPFFdiv2ff(grecip, DUPFFlc(tmp));
  DUPFFdiv2ff(hrecip, DUPFFlc(tmp));
  DUPFFfree(tmp);
  ans->grecip = grecip;
  ans->hrecip = hrecip;
  ans->Ep = DUPFFnew(df);
  ans->fr = DUPInew(df);
  ans->g_lifter = g_lifter;
  ans->h_lifter = h_lifter;
  /* Compute in E the carry, a multiple of p */
  DUPFFmul3(ans->Ep, ans->g[0], ans->h[0]);
  ans->E = DUPFF_to_DUPI(ans->Ep);
  DUPIsubtract_product(ans->E, ans->g[0], ans->h[0]);

  return ans;
}


void DUPZfactor_lift_dtor(DUPZfactor_lifter this)
{
  int i;
  int r;

  if (this == NULL) return;
  r = this->r;
  DUPIfree(this->fr);
  DUPIfree(this->E);
  DUPFFfree(this->Ep);
  if (this->g_lifter) DUPZfactor_lift_dtor(this->g_lifter);
  if (this->h_lifter) DUPZfactor_lift_dtor(this->h_lifter);
  for (i=0; i < r; i++)
  {
    DUPFFfree(this->g[i]);
    DUPFFfree(this->h[i]);
  }
  FREE(this->g);
  FREE(this->h);
  DUPFFfree(this->grecip);
  DUPFFfree(this->hrecip);
  FREE(this);
}

/***************************************************************************/

void DUPZfactor_lift_step(DUPZfactor_lifter this, DUPFF f_cmpt_r)
{
  int i, p;
  /* the rest are just aliases to make the code easier to read (!?!) */
  DUPFF       Ep     = this->Ep;
  DUPI        fr     = this->fr;
  DUPFF       *g     = this->g;
  DUPFF       *h     = this->h;
  int         r      = this->r;
  const DUPFF grecip = this->grecip;
  const DUPFF hrecip = this->hrecip;
  DUPI        E      = this->E;

  p = CurrentFF.prime;
  DUPIdiv2z(E, p);

  DUPFF_to_DUPI2(fr, f_cmpt_r);
  DUPIadd3(E, E, fr);

  for (i=1; i < r; i++)
    DUPIsubtract_product(E, g[i], h[r-i]);

  DUPI_to_DUPFF2(Ep, E);
  solve_for_correction(&g[r], Ep, g[0], hrecip);
  solve_for_correction(&h[r], Ep, h[0], grecip);
  DUPIsubtract_product(E, g[0], h[r]);
  DUPIsubtract_product(E, h[0], g[r]);
  this->r++;

  if (this->g_lifter) DUPZfactor_lift_step(this->g_lifter, g[r]);
  if (this->h_lifter) DUPZfactor_lift_step(this->h_lifter, h[r]);
}


/***************************************************************************/
/* Conversion of poly from its p-adic expansion to normal representation.  */
/* This converts an array of polynomials modulo p to a polynomial over Z.  */
/* Result = array[0] + p*array[1] + p^2*array[2] + ... + p^(k-1)*array[k-1]*/

DUPZ DUPZ_from_padic(const DUPFF *F, int p, int k)
{
  DUPZ ans;
  int i, j, dans;

  dans = DUPFFdeg(F[0]);
  ans = DUPZnew(dans); /* all coeffs initialized to 0 */
  for (i=k-1; i >= 0; i--)
  {
    for (j=0; j <= dans; j++)
    {
      mpz_mul_ui(ans->coeffs[j], ans->coeffs[j], p);
      if (j <= DUPFFdeg(F[i]))
	mpz_add_ui(ans->coeffs[j], ans->coeffs[j], F[i]->coeffs[j]);
    }
  }
  ans->deg = DUPFFdeg(F[0]);
  return ans;
}

/***************************************************************************/
/* Convert the p-adic expansions of the lifted factors to their normal     */
/* representations.                                                        */

int DUPZfactor_lift_output(DUPZ *factors, const DUPZfactor_lifter this, int i)
{
  int p, k;

  p = CurrentFF.prime;
  k = this->r;
  if (this->g_lifter) i = DUPZfactor_lift_output(factors, this->g_lifter, i);
  else factors[i++] = DUPZ_from_padic(this->g, p, k);
  if (this->h_lifter) i = DUPZfactor_lift_output(factors, this->h_lifter, i);
  else factors[i++] = DUPZ_from_padic(this->h, p, k);
  return i;
}



/***************************************************************************/

static DUPZfactor_lifter DUPZfactor_lift_init1(const DUPFFlist pfactors, int rmax)
{
  DUPZfactor_lifter *lifter, ans, new;
  DUPFF **pfac;
  int i, j, nfactors, *deg, newdeg;
  DUPFFlist iter;
#ifdef JAA_FLAG
  DUPFF fi, fj, tmp, *xyz;
  int k, t, tmin, jmin, d;
#endif

  nfactors = DUPFFlist_length(pfactors);
  pfac = (DUPFF**)MALLOC(nfactors*sizeof(DUPFF*));
  deg = (int*)MALLOC(nfactors*sizeof(int));
  for (i=0, iter=pfactors; iter; iter = iter->next, i++)
  {
    pfac[i] = &iter->poly;
    deg[i] = DUPFFdeg(iter->poly);
  }

  lifter = (DUPZfactor_lifter*)MALLOC(nfactors*sizeof(DUPZfactor_lifter));
  for (i=0; i < nfactors; i++) lifter[i] = NULL;
  for (i=0; i < nfactors-1; i++)
  {
#ifdef JAA_FLAG
    if (i+1 == nfactors-1) goto skip;;
    d = deg[i+1];
    tmin = 0;  jmin = -987654321;
    for (j=i+1; j < nfactors; j++)
    {
      if (deg[j] > d) break;
      if (lifter[i]) fi = DUPFFmul(lifter[i]->g[0], lifter[i]->h[0]);
      else fi = pfac[i][0];
      if (lifter[j]) fj = DUPFFmul(lifter[j]->g[0], lifter[j]->h[0]);
      else fj = pfac[j][0];
      tmp = DUPFFmul(fi, fj);
      t = 0;
      for (k=0; k <= DUPFFdeg(tmp); k++) if (tmp->coeffs[k]) t++;
      if (tmin != 0 && t >= tmin) continue;
      tmin = t;
      jmin = j;
    }
    new = lifter[jmin]; newdeg = deg[jmin];
    xyz = pfac[jmin];
    for (j=jmin; j>i+1; j--)
    {
      deg[j] = deg[j-1];
      pfac[j] = pfac[j-1];
      lifter[j] = lifter[j-1];
    }
    deg[i+1] = newdeg;
    lifter[i+1] = new;
    pfac[i+1] = xyz;
  skip:
#endif
    new = DUPZfactor_lifter_ctor(pfac[i], pfac[i+1], lifter[i], lifter[i+1], rmax);
    newdeg = deg[i] + deg[i+1];
#if 0
lifter[i+1]=new;deg[i+1]=newdeg;/* DELETE ME -- BUG BUG BUG BUG BUG */
#else
    /* Re-sort the array; i.e. insert new lifter in correct position */
    for (j=i+2; j < nfactors; j++)
    {
      if (newdeg <= deg[j]) break;
      deg[j-1] = deg[j];
      pfac[j-1] = pfac[j];
      lifter[j-1] = lifter[j];
    }
    lifter[j-1] = new;
    deg[j-1] = newdeg;
#endif
  }
  ans = lifter[nfactors-1];

  FREE(lifter); /* do not free its elements! */
  FREE(pfac);
  FREE(deg);

  return ans;
}


void DUPZfactor_lift_init(DUPZfactor_info this)
{
  const int p = CurrentFF.prime;

  /* We are sure to get a non-trivial factorization before height rmax: */
  this->rmax = 1+ 1.01*(this->bounds->leading_coeff + this->bounds->lift_bound)/logi(p);
  this->lifter = DUPZfactor_lift_init1(this->pfactors, this->rmax);
  this->f_expansion = DUPZpadic_expand_init(this->f, this->p, this->rmax);
}




/****************************************************************************/
/* This function is called after a speculative early search which has found */
/* at least one factor but further lifting is still required.               */
/* It creates a new lifting tree for the remaining factors.                 */

void DUPZfactor_lift_revise1(DUPZfactor_info info, int height)
{
  int i;

  /* now "lift" to the previously attained height */
  /* We do this the stupid way for now... (how to do this better? BUG) */
  DUPZfactor_lift_init(info);
  DUPZpadic_expand_end(info->f_expansion);
  info->f_expansion = DUPZpadic_expand_init(info->f, info->p, info->rmax);
  DUPZpadic_expand_step(info->f_expansion);
  if (height > info->rmax) height = info->rmax;
  for(i=1; i < height; i++)
  {
    DUPZpadic_expand_step(info->f_expansion);
    DUPZfactor_lift_step(info->lifter, info->f_expansion->monic_cmpts[i]);
  }
}


void DUPZfactor_lift_revise(DUPZfactor_info info)
{
  int height = info->lifter->r;

  DUPZfactor_lift_dtor(info->lifter); info->lifter = NULL;
  DUPZpadic_expand_end(info->f_expansion); info->f_expansion = NULL;
  DUPZfactor_lift_revise1(info, height);
}


/***************************************************************************/
void liftprint(DUPZfactor_lifter this, int indent)
{
int i;
if (this == NULL) return;
putchar('>');for(i=0;i<indent;i++)putchar(' ');
for(i=0;i<this->r;i++)printf("%d,",DUPFFdeg(this->g[i]));printf("\n");
for(i=0;i<=indent;i++)putchar(' ');
for(i=0;i<this->r;i++)printf("%d,",DUPFFdeg(this->h[i]));printf("\n");
liftprint(this->g_lifter, indent+2);
liftprint(this->h_lifter, indent+2);
}

#endif
